"""Function to setup axes for barplot plotting."""

from typing import Tuple, Dict, List, Iterable
from math import ceil
from matplotlib.figure import Figure
from matplotlib.axes import Axes
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from sanitize_ml_labels import sanitize_ml_labels
from barplots.utils.get_best_match import get_best_match
from barplots.utils.get_max_bar_position import get_max_bar_position


GOLDEN_RATIO: float = 1.61803398875


def swap(*args: List, flag: bool) -> Iterable:
    """If the given flag is true returns"""
    return args if flag else reversed(args)


def get_axes(
    df: pd.DataFrame,
    bar_width: float,
    space_width: float,
    height: float,
    dpi: int,
    title: str,
    data_label: str,
    vertical: bool,
    subplots: bool,
    titles: List[str],
    plots_per_row: int,
    custom_defaults: Dict[str, List[str]],
    expected_levels: int,
    scale: str,
    sanitize_metrics: bool,
    facecolors: Dict[str, str],
    show_title: bool,
    show_column_name: bool,
) -> Tuple[Figure, Axes]:
    """Setup axes for barplot plotting.

    Parameters
    ----------
    df: pd.DataFrame,
        Dataframe from which to obtain the curresponding barplot width.
    bar_width: float,
        Width of bars in considered barplot.
    space_width: float,
        Width of spaces between spaces.
    height: float,
        Height of considered barplot.
    dpi: int,
        DPI for rendered images.
    title: str,
        Title of the considered barplot.
    data_label: str,
        barplot's data_label. None for not showing any data_label (default).
    vertical: bool,
        Whetever to build the axis to show the bars as vertical or as horizontal.
    expected_levels: int,
        Number of levels to expect to plot as labels.
    scale: str,
        Scale to use for the barplots.
        Can either be "linear" or "log".
    sanitize_metrics: bool,
        Whetever to sanitize the metrics or not.
    facecolors: Dict[str, str],
        Colors for the background.
    show_title: str = True,
        Whetever to show or not the barplot title.
    show_column_name: bool = True
        Whether to show the metric name.

    Returns
    -----------
    Tuple containing new figure and axis.
    """
    if subplots:
        index: MultiIndex = df.index
        side = max(
            get_max_bar_position(df.loc[index], bar_width, space_width)
            for index in index.levels[0]
        )
    else:
        side = get_max_bar_position(df, bar_width, space_width)

    if height is None:
        exponent = 1 if subplots or expected_levels > 1 else 1.5
        height = side / (GOLDEN_RATIO**exponent)

    if subplots:
        nrows = ceil(df.index.levels[0].size / plots_per_row)
    else:
        nrows = plots_per_row = 1

    width, height = swap(side, height, flag=vertical)
    fig, axes = plt.subplots(
        nrows=nrows,
        ncols=plots_per_row,
        figsize=(width * plots_per_row, height * nrows),
        dpi=dpi,
    )
    fig.patch.set_facecolor("white")

    if isinstance(axes, Axes):
        axes = np.array([axes])

    axes = axes.flatten()

    for subtitle, ax in zip(titles, axes):
        ax.set_facecolor(get_best_match(facecolors, subtitle))
        if vertical:
            ax.set_yscale(scale)
            ax.set_xlim(0, side)
            ax.set_xticks([])
            ax.yaxis.grid(True, which="both")
            if data_label is not None and show_column_name:
                ax.set_ylabel(
                    sanitize_ml_labels(data_label, custom_defaults=custom_defaults)
                    if sanitize_metrics
                    else data_label
                )
        else:
            ax.set_xscale(scale)
            ax.set_ylim(0, side)
            ax.set_yticks([])
            ax.xaxis.grid(True, which="both")
            if data_label is not None and show_column_name:
                ax.set_xlabel(
                    sanitize_ml_labels(data_label, custom_defaults=custom_defaults)
                    if sanitize_metrics
                    else data_label
                )
        if show_title:
            ax.set_title(
                sanitize_ml_labels(subtitle, custom_defaults=custom_defaults)
                if sanitize_metrics
                else subtitle
            )

    for ax in axes[len(titles) :]:
        ax.grid(False)
        ax.axis("off")

    if title is not None and len(axes) == 1 and show_title:
        axes[0].set_title(
            sanitize_ml_labels(title, custom_defaults=custom_defaults)
            if sanitize_metrics
            else title
        )

    return fig, axes



"""Module to generate bar positions."""

from typing import Generator, Tuple
import pandas as pd
from barplots.utils.get_jumps import get_jumps


def bar_positions(df: pd.DataFrame, bar_width: float, space_width: float) -> Generator:
    """Returns a generator of bar positions.

    Parameters
    ----------
    df: pd.DataFrame
        Dataframe to iterate to extract the necessary data.
    bar_width: float
        Width of any given bar.
    space_width: float
        Width of spaces between spaces.
    """
    old_index: Tuple = ()
    bar_position: float = 0.0
    for index, values in df.iterrows():
        if not isinstance(index, (list, tuple)):
            index = (index,)

        jumps = get_jumps(index, old_index)

        bar_position += space_width * sum(jumps[:-1])
        if jumps[-1]:
            bar_position += bar_width

        old_index = index

        if len(values) == 2:
            y, std = values
        elif len(values) == 1:
            y, std = values.values[0], 0
        else:
            raise ValueError(
                "Dataframe must have 1 or 2 columns, "
                f"but {len(values)} columns were found."
            )

        yield (bar_position + bar_width / 2, y, std, index)



"""Utility dispatch function to plot bar with given properties."""

from matplotlib.axis import Axis


def plot_bar(
    axes: Axis,
    x: float,
    y: float,
    std: float,
    min_std: float,
    bar_width: float,
    vertical: bool,
    **kwargs
):
    """Plot bar with given properties.

    Parameters
    ----------
    axes: Axis,
        Axis object where to plot the bar.
    x: float,
        Position for the left size of the bar.
    y: float,
        Height of the considered bar.
    std: float,
        Standard deviation to plot on top.
    min_std: float,
        Minimum standard deviation to be shown.
    bar_width: float,
        Width of the bar.
    vertical: bool,
        Whetever to build the axis to show the bars as vertical or as horizontal.
    """
    if vertical:
        axes.bar(
            x=x,
            height=y,
            width=bar_width,
            **({"yerr": std} if std > min_std else {}),
            capsize=7 * bar_width / 0.3,
            **kwargs
        )
    else:
        axes.barh(
            y=x,
            width=y,
            height=bar_width,
            **({"xerr": std} if std > min_std else {}),
            capsize=7 * bar_width / 0.3,
            **kwargs
        )



import pandas as pd
from barplots import barplots
import matplotlib.pyplot as plt


def test_simple_barplots():
    root = "test_barplots"
    df = pd.read_csv("tests/test_case.csv")

    custom_defaults = {
        "P": "promoters",
        "E": "enhancers",
        "A": "active ",
        "I": "inactive ",
        "+": " and ",
        "": "anything",
        "Validation": "val",
    }

    barplots(
        df,
        ["cell_line", "task", "model"],
        path="{root}/{{feature}}.png".format(root=root),
        orientation="horizontal",
        verbose=False,
        custom_defaults=custom_defaults,
        subplots=True,
        space_width=1,
    )
    plt.close()



from barplots.utils.bar_positions import bar_positions
import pandas as pd


def execute_test(df, ground_truth):
    df = df.groupby(["cell_line", "task", "balancing", "model"]).mean()

    for x1, x2 in zip(ground_truth, bar_positions(df, 0.5, 0.5)):
        assert x1 == x2[0]


def test_bar_positions():
    df1 = pd.read_csv("tests/test_case.csv")
    df1 = df1[df1.cell_line == "HelaS3"]
    df1 = df1[df1.balancing == "unbalanced"]
    df2 = df1[df1.model.str.contains("bayesian").values]
    df3 = df2[df2.model.str.contains("mlp").values]

    execute_test(
        df1,
        [
            0.25,
            0.75,
            1.25,
            1.75,
            3.25,
            3.75,
            4.25,
            4.75,
            6.25,
            6.75,
            7.25,
            7.75,
            9.25,
            9.75,
            10.25,
            10.75,
            12.25,
            12.75,
            13.25,
            13.75,
        ],
    )
    execute_test(df2, [0.25, 0.75, 2.25, 2.75, 4.25, 4.75, 6.25, 6.75, 8.25, 8.75])
    execute_test(df3, [0.25, 1.25, 2.25, 3.25, 4.25])



import pandas as pd
import pytest
from barplots import barplots
import warnings


def test_wrong_parameters():
    root = "test_barplots"
    df = pd.read_csv("tests/test_case.csv")

    with pytest.raises(ValueError):
        barplots(
            df[df.cell_line == "CIAO"],
            ["cell_line", "task", "balancing", "model"],
            path="{root}/{{feature}}.png".format(root=root),
        )

    with pytest.raises(ValueError):
        barplots(
            df,
            ["cell_line", "task", "balancing", "model"],
            orientation="pinco",
            path="{root}/{{feature}}.png".format(root=root),
        )

    with pytest.raises(ValueError):
        barplots(
            df,
            ["model"],
            path="{root}/{{feature}}.png".format(root=root),
            subplots=True,
        )

    with pytest.raises(ValueError):
        barplots(
            df,
            ["model"],
            plots_per_row="pinco",
            path="{root}/{{feature}}.png".format(root=root),
            subplots=True,
        )



from typing import Union, List, Any


def get_jumps(index: Union[List, Any], old_index: Union[List, Any]) -> List[bool]:
    """Return list representing the detected jumps from given index and old_index.

    Parameters
    ----------
    index:Union[List, Any],
        List of indices of index curresponding to given row number.
    old_index:Union[List, Any],
        Old index.

    Returns
    -------
    Returns list of boolean representing if for given index level a jump has been detected.
    """
    if not old_index:
        return [False] * len(index)

    previous = False
    jumps = []
    for new, old in zip(index, old_index):
        jumps.append(new != old or previous)
        previous = new != old
    return jumps



"""Plot bars for given dataframe at given intervals."""

from typing import Dict, Optional
import pandas as pd
from matplotlib.axes import Axes
from barplots.utils.plot_bar import plot_bar
from barplots.utils.bar_positions import bar_positions
from barplots.utils.get_best_match import get_best_match


def plot_bars(
    axes: Axes,
    df: pd.DataFrame,
    bar_width: float,
    space_width: float,
    alphas: Dict[str, float],
    infer_alphas: bool,
    colors: Dict[str, str],
    infer_colors: bool,
    edgecolors: Optional[Dict[str, str]],
    infer_edgecolors: bool,
    hatch: Optional[Dict[str, str]],
    infer_hatch: bool,
    top_index: str,
    **kwargs: Dict
):
    """Plot bars for given dataframe at given intervals.

    Parameters
    ----------
    axes:Axes,
        The axes where to plot the bars.
    df: pd.DataFrame,
        The dataframe from where to extract the data.
    bar_width: float,
        The width of the bars, used also for spacing.
    space_width: float,
        Width of spaces between spaces.
    alphas: Dict[str, float],
        Dictionary of alphas to be used.
    infer_alphas: bool,
        Whetever to infer alphas or not.
    colors: Dict[str, str],
        Dictionary of colors to be used.
    infer_colors: bool,
        Whetever to infer colors or not.
    edgecolors: Optional[Dict[str, str]],
        Dictionary of edgecolors to be used.
    infer_edgecolors: bool,
        Whetever to infer edgecolors or not.
    hatch: Optional[Dict[str, str]]
        Dict of hatch, i.e. patterns for the bars, to be used for innermost index of dataframe.
    infer_hatch: bool,
        Whetever to infer hatch or not.
    kwargs: Dict,
        Parameters to be passed directly to the plot_bar method
    """
    for x, y, std, index in bar_positions(df, bar_width, space_width):
        plot_bar(
            axes=axes,
            x=x,
            y=y,
            std=std,
            bar_width=bar_width,
            alpha=(
                alphas[index[-1]]
                if index[-1] in alphas
                else (
                    get_best_match(alphas, (top_index, *index))
                    if infer_alphas
                    else None
                )
            ),
            color=(
                colors[index[-1]]
                if index[-1] in colors
                else (
                    get_best_match(colors, (top_index, *index))
                    if infer_colors
                    else None
                )
            ),
            edgecolor=(
                None
                if edgecolors is None
                else (
                    edgecolors[index[-1]]
                    if index[-1] in edgecolors
                    else (
                        get_best_match(edgecolors, (top_index, *index))
                        if infer_edgecolors
                        else None
                    )
                )
            ),
            hatch=(
                None
                if hatch is None
                else (
                    hatch[index[-1]]
                    if index[-1] in hatch
                    else (
                        get_best_match(hatch, (top_index, *index))
                        if infer_hatch
                        else None
                    )
                )
            ),
            label=index[-1],
            **kwargs
        )



"""Function to get the maximum and minimum bar length, including std."""

from typing import Tuple
import pandas as pd
from barplots.utils.bar_positions import bar_positions


def get_max_bar_length(
    df: pd.DataFrame, bar_width: float, space_width: float
) -> Tuple[float, float]:
    """Return Tuple containing maximum and minimum bar length, including std.

    These values could also be negative.

    Parameters
    ----------
    df: pd.DataFrame
        The dataframe from where to extract the data.
    bar_width: float
        The width of the bars, used also for spacing
    space_width: float
        Width of spaces between spaces.
    """
    return max(
        y + std for _, y, std, _ in bar_positions(df, bar_width, space_width)
    ), min(y - std for _, y, std, _ in bar_positions(df, bar_width, space_width))



"""Utility function to extract index levels from a dataframe."""

from typing import List, Any
import pandas as pd
from pandas import MultiIndex


def get_levels(df: pd.DataFrame) -> List[List[Any]]:
    """Return normalized list of dataframe index levels.

    Parameters
    ----------
    df: pd.DataFrame,
        Dataframe from which to extract index levels.

    Returns
    -------
    List of lists of unique indices.
    """
    if isinstance(df.index, MultiIndex):
        return [list(e) for e in df.index.levels]
    return [df.index.unique()]



import pandas as pd
from barplots import barplots
import os
import itertools
from tqdm.auto import tqdm
import matplotlib.pyplot as plt


def test_barplots():
    df1 = pd.read_csv("tests/test_case.csv")
    df2 = df1[df1.model.str.contains("bayesian")]
    fuzzy_args = {
        "df": [(1, df1), (2, df2)],
        "groupby": [["cell_line", "task", "model"]],
        "show_legend": [True, False],
        "orientation": ["horizontal", "vertical"],
        "major_rotation": [0, 90],
        "minor_rotation": [0, 90],
        "plots_per_row": [3, "auto"],
        "unique_minor_labels": [True, False],
        "unique_major_labels": [True, False],
        "unique_data_label": [False, True],
        "subplots": [True, False],
    }

    custom_defaults = {
        "P": "promoters",
        "E": "enhancers",
        "A": "active ",
        "I": "inactive ",
        "+": " and ",
        "": "anything",
        "Validation": "val",
    }

    arguments = list(itertools.product(*list(fuzzy_args.values())))
    for arg in tqdm(arguments, desc="Running test suite"):
        kwargs = dict(zip(fuzzy_args.keys(), arg))
        path = "examples/{i}/{orientation}".format(
            i=kwargs["df"][0], orientation=kwargs["orientation"]
        )

        os.makedirs(os.path.dirname(path), exist_ok=True)

        if kwargs["show_legend"]:
            path += "_legend"

        if kwargs["subplots"]:
            path += "_subplots"
        else:
            kwargs["groupby"] = kwargs["groupby"][1:]

        if kwargs["show_legend"] and kwargs["major_rotation"]:
            continue

        if (
            kwargs["show_legend"]
            and kwargs["orientation"] == "horizontal"
            and not kwargs["minor_rotation"]
        ):
            continue

        if (
            kwargs["show_legend"]
            and kwargs["orientation"] == "vertical"
            and kwargs["minor_rotation"]
        ):
            continue

        if (
            not kwargs["show_legend"]
            and kwargs["orientation"] == "vertical"
            and not kwargs["minor_rotation"]
        ):
            continue

        if (
            not kwargs["show_legend"]
            and kwargs["orientation"] == "horizontal"
            and kwargs["minor_rotation"]
        ):
            continue

        if (
            not kwargs["show_legend"]
            and kwargs["orientation"] == "vertical"
            and kwargs["major_rotation"]
        ):
            continue

        if (
            not (kwargs["minor_rotation"] or kwargs["major_rotation"])
            and kwargs["orientation"] == "horizontal"
        ):
            continue

        if (kwargs["show_legend"] or not kwargs["subplots"]) and (
            kwargs["unique_data_label"]
            or kwargs["unique_minor_labels"]
            or kwargs["unique_major_labels"]
        ):
            continue

        if kwargs["major_rotation"]:
            path += "_major_rotation"

        if kwargs["minor_rotation"]:
            path += "_minor_rotation"

        if kwargs["unique_minor_labels"]:
            path += "_unique_minor_labels"

        if kwargs["unique_major_labels"]:
            path += "_unique_major_labels"

        path += "_{feature}.png"

        kwargs["df"] = kwargs["df"][1]

        barplots(**kwargs, path=path, custom_defaults=custom_defaults, verbose=False)
        plt.close()


def test_single_index():
    root = "test_barplots"
    df = pd.read_csv("tests/test_case.csv")
    barplots(
        df,
        ["cell_line"],
        path="{root}/{{feature}}.png".format(root=root),
        verbose=False,
    )
    plt.close()



"""Remove duplicated labels from the plot legend."""

from typing import Dict, List, Optional
import math
from matplotlib.axes import Axes
from matplotlib.patches import Patch
from sanitize_ml_labels import sanitize_ml_labels


def remove_duplicated_legend_labels(
    axes: Axes,
    legend_position: str,
    legend_title: str,
    legend_entries_size: float,
    legend_title_size: float,
    show_legend_title: bool,
    sanitize_labels: bool,
    custom_defaults: Dict[str, List[str]],
    ncol: Optional[int] = None,
):
    """Remove duplicated labels from the plot legend.

    Parameters
    ----------
    axes: Axes
        Axes where to show the labels.
    legend_position: str
        Legend position.
    legend_title: str
        Title for the legend.
    legend_entries_size: float
        Size for the legend entries font.
    legend_title_size: float
        Size for the legend title font.
    show_legend_title: bool
        Whether to show the legend title.
    sanitize_labels: bool
        Whether to sanitize the labels or not.
    custom_defaults: Dict[str, List[str]]
        The defaults for normalizing the provided keys.
    ncol: Optional[int] = None
        The number of columns to show in the barplot.
    """
    handles, labels = axes.get_legend_handles_labels()

    by_label = dict(zip(labels, handles))
    length__of_padding = 6
    mean_label_length = (
        sum(len(label) for label in by_label.keys()) / len(by_label)
        + length__of_padding
    )

    ncol = math.ceil(len(legend_title) / mean_label_length) if ncol is None else ncol

    legend = axes.legend(
        handles=[
            Patch(
                linestyle="none",
                label=label,
                linewidth=legend_entries_size,
                facecolor=handler.patches[0].get_facecolor(),
                hatch=handler.patches[0].get_hatch(),
            )
            for handler, label in zip(
                by_label.values(),
                (
                    sanitize_ml_labels(by_label.keys(), custom_defaults=custom_defaults)
                    if sanitize_labels
                    else by_label.keys()
                ),
            )
        ],
        ncol=ncol,
        handletextpad=0.1,
        columnspacing=0.1,
        handlelength=0.7,
        prop={"size": legend_entries_size},
        loc=legend_position,
    )
    if show_legend_title:
        legend.set_title(
            (
                sanitize_ml_labels(legend_title, custom_defaults=custom_defaults)
                if sanitize_labels
                else legend_title
            ),
            prop={"weight": "bold", "size": legend_title_size},
        )



"""Submodule handling the plotting of the bar labels."""

from typing import Dict, List, Union, Optional

import pandas as pd
from matplotlib.axes import Axes
from matplotlib.figure import Figure
import matplotlib.pyplot as plt
from sanitize_ml_labels import sanitize_ml_labels

from barplots.utils.get_max_bar_position import get_max_bar_position
from barplots.utils.text_positions import text_positions

factors = [
    ("_", 0),
    ("y", 1e-24),
    ("z", 1e-21),
    ("a", 1e-18),
    ("f", 1e-15),
    ("p", 1e-12),
    ("n", 1e-9),
    ("µ", 1e-6),
    ("m", 1e-3),
    ("K", 1e3),
    ("M", 1e6),
    ("G", 1e9),
    ("T", 1e12),
    ("P", 1e15),
    ("E", 1e18),
    ("Z", 1e21),
    ("Y", 1e24),
    ("_", float("inf")),
]


def sanitize_digits(digit: float, unit: Optional[str], normalized: bool):
    unit = "" if unit is None else unit
    absolute_digit = abs(digit)
    if (
        not normalized
        and digit != 0.0
        and (absolute_digit <= 1e-3 or absolute_digit >= 1e3)
    ):
        for (lower_factor, lower_value), (higher_factor, higher_value) in zip(
            factors[:-1],
            factors[1:],
        ):
            if absolute_digit < higher_value:
                if lower_factor == "_":
                    denominator = higher_value
                    factor = higher_factor
                else:
                    denominator = lower_value
                    factor = lower_factor
                digit /= denominator
                unit = factor + unit
                break

    return sanitize_ml_labels(digit) + unit


def plot_bar_labels(
    axes: Axes,
    figure: Figure,
    df: pd.DataFrame,
    vertical: bool,
    levels: int,
    bar_width: float,
    space_width: float,
    minor_rotation: Union[float, str],
    major_rotation: Union[float, str],
    unique_minor_labels: bool,
    unique_major_labels: bool,
    unique_data_label: bool,
    custom_defaults: Dict[str, List[str]],
    unit: Optional[str],
    normalized_metric: bool,
    absolutely_normalized_metric: bool,
    sanitize_metrics: bool,
):
    """
    Parameters
    ------------
    minor_rotation: Union[float, str]
        Rotation for the minor ticks of the bars.
        By default, with the "auto" mode, the library tries to find
        the rotation with which we minimize the overlap for the provided
        labels, including also the overlap with minor and major.
    major_rotation: Union[float, str]
        Rotation for the major ticks of the bars.
        By default, with the "auto" mode, the library tries to find
        the rotation with which we minimize the overlap for the provided
        labels, including also the overlap with minor and major.
    unique_minor_labels: bool = True,
        Avoid replicating minor labels on the same axis in multiple subplots settings.
    unique_major_labels: bool = True,
        Avoid replicating major labels on the same axis in multiple subplots settings.
    unique_data_label: bool = True,
        Avoid replication of data axis label when using subplots.
    unit: Optional[str]
        Optional unit to show on the value axis.
    normalized_metric: bool
        Whether to consider the current metric normalized in a range (0, 1)
    absolutely_normalized_metric: bool
        Whether to consider the current metric absolutely normalized in a range (-1, 1)
    sanitize_metrics: bool
        Whether to sanitize the metrics or not.
    """
    other_positions = set()
    width = get_max_bar_position(df, bar_width, space_width)

    if unique_data_label:
        axes.set_ylabel("")

    if normalized_metric or absolutely_normalized_metric:
        nbins = 5 if normalized_metric else 8

        if vertical:
            axes.locator_params(axis="y", nbins=nbins)
        else:
            axes.locator_params(axis="x", nbins=nbins)

    def sanitizer(x, *args, **kwargs):
        return sanitize_digits(
            x, unit=unit, normalized=normalized_metric or absolutely_normalized_metric
        )

    if vertical:
        axes.yaxis.set_major_formatter(plt.FuncFormatter(sanitizer))
    else:
        axes.xaxis.set_major_formatter(plt.FuncFormatter(sanitizer))

    for level in reversed(range(max(levels - 2, 0), levels)):
        positions, labels = zip(*text_positions(df, bar_width, space_width, level))
        labels = (
            sanitize_ml_labels(labels, custom_defaults=custom_defaults)
            if sanitize_metrics
            else labels
        )

        max_characters_number_in_labels = max((len(label) for label in labels))

        positions = [round(pos, 5) for pos in positions]
        positions = [
            position + width * 0.0002 if position in other_positions else position
            for position in positions
        ]
        other_positions |= set(positions)
        minor = level == levels - 1

        # Handle the automatic rotation of minor labels.
        if minor_rotation == "auto":
            if (
                minor
                and len(set(labels)) <= width * 5 / max_characters_number_in_labels
                and vertical
            ):
                adapted_minor_rotation = 90
            elif (
                minor
                and len(set(labels)) <= width * 20 / max_characters_number_in_labels
                and not vertical
            ):
                adapted_minor_rotation = 90
            else:
                adapted_minor_rotation = 0
        else:
            adapted_minor_rotation = minor_rotation

        # Handle the automatic rotation of major labels.
        if major_rotation == "auto":
            if (
                not minor
                and len(set(labels)) <= width * 5 / max_characters_number_in_labels
                and not vertical
            ):
                adapted_major_rotation = 90
            elif (
                not minor
                and len(set(labels)) >= width * 20 / max_characters_number_in_labels
                and vertical
            ):
                adapted_major_rotation = 90
            else:
                adapted_major_rotation = 0
        else:
            adapted_major_rotation = major_rotation

        if minor and unique_minor_labels:
            continue
        if not minor and unique_major_labels:
            continue

        if vertical:
            axes.set_xticks(positions, minor=minor)
            axes.set_xticklabels(labels, minor=minor, ha="center")
            if minor:
                axes.tick_params(
                    axis="x",
                    labelsize=9,
                    which="minor",
                    labelrotation=adapted_minor_rotation,
                )

                if adapted_minor_rotation > 80:
                    length = 6 * (max_characters_number_in_labels + 1)
                else:
                    length = 20

                axes.tick_params(
                    axis="x",
                    labelsize=10,
                    which="major",
                    direction="out",
                    length=length,
                    width=0,
                )
            else:
                axes.tick_params(
                    axis="x",
                    labelsize=10,
                    which="major",
                    labelrotation=adapted_major_rotation,
                )
        else:
            axes.set_yticks(positions, minor=minor)
            axes.set_yticklabels(labels, minor=minor, va="center")
            if minor:
                axes.tick_params(
                    axis="y",
                    which="minor",
                    labelsize=9,
                    labelrotation=adapted_minor_rotation,
                )

                if adapted_minor_rotation > 80:
                    length = 20
                else:
                    length = 6 * (max_characters_number_in_labels + 1)

                axes.tick_params(
                    axis="y",
                    which="major",
                    labelsize=10,
                    direction="out",
                    length=length,
                    # This is the size of the actual `tick`
                    # in the plot, which we do not want to show
                    # for the major ticks in this case and therefore
                    # we set it to zero.
                    width=0,
                )
            else:
                axes.tick_params(
                    axis="y", which="major", labelrotation=adapted_major_rotation
                )



"""Save the given figure to the given path."""

import os
from matplotlib.figure import Figure


def save_picture(path: str, figure: Figure):
    """Save the given figure to the given path.

    Parameters
    ----------
    path: str,
        Path where to save the figure.
    figure: Figure,
        Figure to save.
    """
    directory = os.path.dirname(path)
    if directory:
        os.makedirs(directory, exist_ok=True)
    figure.savefig(path, bbox_inches="tight")



"""Module implementing plotting of a barplot."""

from typing import List, Tuple, Dict, Union, Callable, Optional
import pandas as pd
from matplotlib.figure import Figure
from matplotlib.axes import Axes
from sanitize_ml_labels import is_normalized_metric, is_absolutely_normalized_metric
from barplots.utils import (
    get_axes,
    get_levels,
    remove_duplicated_legend_labels,
    get_max_bar_length,
    save_picture,
    plot_bars,
    plot_bar_labels,
)

# List of 10 distinct colors from the Tableau palette.
TABLEAU_COLORS = [
    "#4e79a7",
    "#f28e2b",
    "#e15759",
    "#76b7b2",
    "#59a14e",
    "#edc949",
    "#b07aa2",
    "#ff9da7",
    "#9c755f",
    "#bab0ac",
]

# List of 20 distinct colors composed by Sasha Trubetskoy.
SASHA_COLORS = [
    "#e6194B",
    "#3cb44b",
    "#ffe119",
    "#4363d8",
    "#f58231",
    "#911eb4",
    "#42d4f4",
    "#f032e6",
    "#bfef45",
    "#fabed4",
    "#469990",
    "#dcbeff",
    "#9A6324",
    "#fffac8",
    "#800000",
    "#aaffc3",
    "#808000",
    "#ffd8b1",
    "#000075",
    "#a9a9a9",
]

# List of hatches supported by matplotlib.
HATCHES = [
    "x",
    "o",
    "*",
    "/",
    "O",
    "\\",
    "|",
    ".",
    "-",
    "+",
]


def barplot(
    df: pd.DataFrame,
    bar_width: float = 0.3,
    space_width: float = 0.2,
    height: Optional[float] = None,
    dpi: int = 200,
    min_std: float = 0,
    min_value: Optional[float] = None,
    max_value: Optional[float] = None,
    show_legend: bool = True,
    show_last_level_as_legend: bool = True,
    show_title: str = True,
    show_column_name: bool = True,
    legend_position: str = "best",
    data_label: Optional[str] = None,
    title: Optional[str] = None,
    path: Optional[str] = None,
    colors: Optional[Dict[str, str]] = None,
    hatch: Optional[Dict[str, str]] = None,
    alphas: Optional[Dict[str, float]] = None,
    facecolors: Optional[Dict[str, str]] = None,
    orientation: str = "vertical",
    subplots: bool = False,
    plots_per_row: Union[int, str] = "auto",
    minor_rotation: Union[float, str] = "auto",
    major_rotation: Union[float, str] = "auto",
    unique_minor_labels: bool = False,
    unique_major_labels: bool = True,
    unique_data_label: bool = True,
    auto_normalize_metrics: bool = True,
    placeholder: bool = False,
    scale: str = "linear",
    sanitize_metrics: bool = True,
    unit: Optional[str] = None,
    legend_entries_size: float = 8,
    legend_title_size: float = 9,
    letter_per_subplot: Optional[List[str]] = None,
    show_legend_title: bool = True,
    custom_defaults: Optional[Dict[str, List[str]]] = None,
    sort_bars: Optional[Callable[[pd.DataFrame], pd.DataFrame]] = None,
    letter: Optional[str] = None,
    letter_font_size: int = 20,
    ncol: Optional[int] = None,
) -> Tuple[Figure, Axes]:
    """Plot barplot corresponding to given dataframe, containing y value and optionally std.

    Parameters
    ----------
    df: pd.DataFrame,
        Dataframe from which to extrat data for plotting barplot.
    bar_width: float = 0.3,
        Width of the bar of the barplot.
    space_width: float = 0.2
        Width of the space between bar groups.
    height: Optional[float] = None,
        Height of the barplot. By default golden ratio of the width.
    dpi: int = 200,
        DPI for plotting the barplots.
    min_std: float = 0.001,
        Minimum standard deviation for showing error bars.
    min_value: Optional[float] = None,
        Minimum value for the barplot.
    max_value: float = 0,
        Maximum value for the barplot.
    show_legend: bool = True,
        Whether to show the legend.
    show_last_level_as_legend: bool = True,
        Whetever to show or not the legend.
        If legend is hidden, the bar ticks are shown alternatively.
    show_title: str = True
        Whetever to show or not the barplot title.
    show_column_name: bool = True
        Whether to show the metric name.
    legend_position: str = "best",
        Legend position, by default "best".
    data_label: Optional[str] = None,
        Barplot's data_label.
        Use None for not showing any data_label (default).
    title: Optional[str] = None,
        Barplot's title.
        Use None for not showing any title (default).
    path: Optional[str] = None,
        Path where to save the barplot.
        Use None for not saving it (default).
    colors: Optional[Dict[str, str]] = None,
        Dict of colors to be used for innermost index of dataframe.
        By default None, using the default color tableau from matplotlib.
    hatch: Optional[Dict[str, str]] = None
        Dict of hatch, i.e. patterns for the bars, to be used for innermost index of dataframe.
    alphas: Optional[Dict[str, float]] = None,
        Dict of alphas to be used for innermost index of dataframe.
        By default None, using the default alpha.
    orientation: str = "vertical",
        Orientation of the bars.
        Can either be "vertical" of "horizontal".
    subplots: bool = False,
        Whetever to slit the top indexing layer to multiple subplots.
    plots_per_row: Union[int, str] = "auto",
        If subplots is True, specifies the number of plots for row.
        If "auto" is used, for vertical the default is 2 plots per row,
        while for horizontal the default is 4 plots per row.
    minor_rotation: Union[float, str] = "auto"
        Rotation for the minor ticks of the bars.
        By default, with the "auto" mode, the library tries to find
        the rotation with which we minimize the overlap for the provided
        labels, including also the overlap with minor and major.
    major_rotation: Union[float, str] = "auto"
        Rotation for the major ticks of the bars.
        By default, with the "auto" mode, the library tries to find
        the rotation with which we minimize the overlap for the provided
        labels, including also the overlap with minor and major.
    unique_minor_labels: bool = False,
        Avoid replicating minor labels on the same axis in multiple subplots settings.
    unique_major_labels: bool = True,
        Avoid replicating major labels on the same axis in multiple subplots settings.
    unique_data_label: bool = True,
        Avoid replication of data axis label when using subplots.
    auto_normalize_metrics: bool = True,
        Whetever to apply or not automatic normalization
        to the metrics that are recognized to be between
        zero and one. For example AUROC, AUPRC or accuracy.
    placeholder: bool = False,
        Whetever to add a text on top of the barplots to show
        the word "placeholder". Useful when generating placeholder data.
    scale: str = "linear",
        Scale to use for the barplots.
        Can either be "linear" or "log".
    sanitize_metrics: bool = True,
        Whetever to sanitize the metrics names.
    unit: Optional[str] = None
        The unit to show in the value axis of the plot.
    legend_entries_size: float = 8
        Size for the legend entries font.
    legend_title_size: float = 9
        Size for the legend title font.
    show_legend_title: bool = True
        Whether to show the legend title.
    custom_defaults: Dict[str, List[str]],
        Dictionary to normalize labels.
    letter: Optional[str] = None
        Letter to show on the top left of the figure.
        This is sometimes necessary on papers.
        By default it is None, that is no letter to be shown.
    letter_per_subplot: Optional[List[str]] = None
        Letter to show on the top left of each subplot.
        This is sometimes necessary on papers.
        By default it is None, that is no letter to be shown.
    letter_font_size: int = 20
        Font size to use for the barplot letter,
        if provided.
    ncol: Optional[int] = None
        The number of columns to show in the barplot.

    Raises
    ------
    ValueError:
        If the given orientation is nor "vertical" nor "horizontal".
    ValueError:
        If the given plots_per_row is nor "auto" or a positive integer.
    ValueError:
        If subplots is True and less than a single index level is provided.

    Returns
    -------
    Tuple containing Figure and Axes of created barplot.
    """

    if orientation not in ("vertical", "horizontal"):
        raise ValueError(f'Given orientation "{orientation}" is not supported.')

    if (
        not isinstance(plots_per_row, int)
        and plots_per_row != "auto"
        or isinstance(plots_per_row, int)
        and plots_per_row < 1
    ):
        raise ValueError(
            f"Given plots_per_row \"{plots_per_row}\" is not 'auto' or a positive integer."
        )

    vertical = orientation == "vertical"

    levels = get_levels(df)
    expected_levels = len(levels) - int(show_last_level_as_legend) - int(subplots)

    if len(levels) <= 1 and subplots:
        raise ValueError("Unable to split plots with only a single index level.")

    if plots_per_row == "auto":
        if subplots:
            plots_per_row = min(
                len(levels[0]), (1 if df.shape[0] > 40 else 2) if vertical else 4
            )
    else:
        plots_per_row = min(plots_per_row, len(levels[0]))

    infer_alphas = alphas is not None
    infer_colors = colors is not None
    infer_hatch = hatch is not None
    infer_edgecolors = False

    edgecolors = None

    if colors is None:
        # When the number of provide faces is less than the
        # tableau colors, we use the tableau colors, else we use
        # the Sasha colors. If even these are not enough, we use
        # the hatches so that we can differentiate the bars more
        # easily.
        if len(levels[-1]) <= len(TABLEAU_COLORS):
            colors = dict(zip(levels[-1], TABLEAU_COLORS))
        elif len(levels[-1]) <= len(SASHA_COLORS):
            colors = dict(zip(levels[-1], SASHA_COLORS))
        else:
            colors: Dict[str, str] = {}
            hatch: Dict[str, str] = {}
            edgecolors: Dict[str, str] = {}
            for i, level in enumerate(levels[-1]):
                colors[level] = TABLEAU_COLORS[i % len(TABLEAU_COLORS)]
                if i >= len(TABLEAU_COLORS):
                    adjusted_i = i // len(TABLEAU_COLORS) - 1
                    hatch[level] = HATCHES[adjusted_i % len(HATCHES)]
                    edgecolors[level] = "white"

    if alphas is None:
        alphas = dict(zip(levels[-1], (0.95,) * len(levels[-1])))

    if facecolors is None:
        facecolors = dict(zip(levels[0], ("white",) * len(levels[0])))

    sorted_level = levels[0]

    if subplots:
        titles = sorted_level
    else:
        titles = ("",)

    figure, axes = get_axes(
        df,
        bar_width,
        space_width,
        height,
        dpi,
        title,
        data_label,
        vertical,
        subplots,
        titles,
        plots_per_row,
        custom_defaults,
        expected_levels,
        scale,
        sanitize_metrics,
        facecolors,
        show_title,
        show_column_name,
    )

    if letter_per_subplot is None:
        letter_per_subplot = ["" for _ in range(len(axes))]

    for i, (subplot_letter, index, ax) in enumerate(
        zip(letter_per_subplot, titles, axes)
    ):
        if subplots:
            sub_df = df.loc[index]
        else:
            sub_df = df

        if sort_bars is not None:
            sub_df = sort_bars(sub_df)

        plot_bars(
            ax,
            sub_df,
            bar_width,
            space_width,
            alphas,
            infer_alphas,
            colors,
            infer_colors,
            edgecolors,
            infer_edgecolors,
            hatch,
            infer_hatch,
            index,
            vertical=vertical,
            min_std=min_std,
        )

        is_not_first_ax = subplots and (
            (not vertical and i % plots_per_row)
            or (vertical and i < len(axes) - plots_per_row)
        )

        is_not_first_vertical_ax = subplots and (
            (vertical and i % plots_per_row)
            or (not vertical and i < len(axes) - plots_per_row)
        )

        normalized_metric = auto_normalize_metrics and (
            is_normalized_metric(df.columns[0][0]) or is_normalized_metric(title)
        )
        absolutely_normalized_metric = auto_normalize_metrics and (
            is_absolutely_normalized_metric(df.columns[0][0])
            or is_absolutely_normalized_metric(title)
        )

        plot_bar_labels(
            ax,
            figure,
            sub_df,
            vertical,
            expected_levels,
            bar_width,
            space_width,
            minor_rotation,
            major_rotation,
            unique_minor_labels and is_not_first_ax,
            unique_major_labels and is_not_first_ax,
            unique_data_label and is_not_first_vertical_ax,
            custom_defaults,
            unit,
            normalized_metric=normalized_metric,
            absolutely_normalized_metric=absolutely_normalized_metric,
            sanitize_metrics=sanitize_metrics,
        )

        ax.text(
            x=-0.1,
            y=1.1,
            s=subplot_letter,
            size=12,
            color="black",
            weight="bold",
            horizontalalignment="left",
            verticalalignment="center",
            transform=ax.transAxes,
        )

        if show_last_level_as_legend and show_legend:
            remove_duplicated_legend_labels(
                ax,
                legend_position,
                df.index.names[-1],
                legend_entries_size,
                legend_title_size,
                show_legend_title,
                custom_defaults,
                ncol,
            )

        max_length, min_length = get_max_bar_length(sub_df, bar_width, space_width)
        max_length *= 1.01
        min_length *= 1.01
        min_length = min(min_length, 0)

        if min_value is not None:
            min_length = min_value

        if normalized_metric:
            max_length = max(max_length, 1.01)
        elif absolutely_normalized_metric:
            max_length = max(max_length, 1.01)
            if min_length < 0:
                min_length = min(min_length, -1.01)

        if max_value is not None:
            max_length = max_value

        if placeholder:
            ax.text(
                0.5,
                0.5,
                "PLACEHOLDER",
                fontsize=30,
                alpha=0.75,
                color="red",
                rotation=8,
                horizontalalignment="center",
                verticalalignment="center",
                transform=ax.transAxes,
            )

        if vertical:
            ax.set_ylim(min_length, max_length)
        else:
            ax.set_xlim(min_length, max_length)

    figure.tight_layout()

    if letter:
        figure.text(
            0,
            0.95,
            letter,
            horizontalalignment="left",
            verticalalignment="top",
            weight="bold",
            fontsize=letter_font_size,
        )

    if path is not None:
        save_picture(path, figure)

    return figure, axes



"""Submodule with utilities for plotting barplots."""

from barplots.utils.save_picture import save_picture
from barplots.utils.get_axes import get_axes
from barplots.utils.text_positions import text_positions
from barplots.utils.plot_bars import plot_bars
from barplots.utils.get_levels import get_levels
from barplots.utils.remove_duplicated_legend_labels import (
    remove_duplicated_legend_labels,
)
from barplots.utils.plot_bar_labels import plot_bar_labels
from barplots.utils.get_max_bar_length import get_max_bar_length


__all__ = [
    "save_picture",
    "get_axes",
    "text_positions",
    "plot_bars",
    "get_levels",
    "remove_duplicated_legend_labels",
    "plot_bar_labels",
    "get_max_bar_length",
]



import pandas as pd
from barplots import barplots
import matplotlib.pyplot as plt


def test_single_index():
    root = "test_barplots"
    df = pd.read_csv("tests/test_case.csv")
    barplots(
        df, "cell_line", path="{root}/{{feature}}.png".format(root=root), verbose=False
    )
    plt.close()



from matplotlib.axis import Axis
import numpy as np


def plot_boxplot(
    axes: Axis,
    data: np.ndarray,
    x: float,
    y: float,
    boxplot_width: float,
    vertical: bool,
    **kwargs
):
    """Plot boxplot with given properties.

    Parameters
    ----------
    axes: Axes,
        Axes object where to plot the boxplot.
    x: float,
        Position for the left size of the boxplot.
    y: float,
        Height of the considered boxplot.
    boxplot_width: float,
        Width of the boxplot.
    vertical: bool,
        Whetever to build the axis to show the boxplots as vertical or as horizontal.
    """
    assert isinstance(x, float)
    assert isinstance(y, float)
    assert isinstance(data, np.ndarray)
    assert np.isfinite(data).all()
    assert isinstance(boxplot_width, float)
    assert isinstance(vertical, bool)

    if vertical:
        axes.boxplot(
            x=data,
            notch=True,
            positions=[x],
            widths=boxplot_width,
            capsize=7 * boxplot_width / 0.3,
            **kwargs
        )
    else:
        raise NotImplementedError("POI FAREMO")
        # axes.boxplot(
        #     y=x,
        #     notch=True,
        #     vert=False,
        #     positions=[x],
        #     widths=y,
        #     height=boxplot_width,
        #     **({"xerr": std} if std > min_std else {}),
        #     capsize=7 * boxplot_width / 0.3,
        #     **kwargs
        # )
